/*
 * Demo write port C for DIL/NetPC PNP/2110-3V
 * Author  : MHA@SSV <info@dilnetpc.com> <m.hasewinkel@web.de>
 * Date    : 29.10.2003
 * Version : 1.00
 * License : GPL
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>

typedef unsigned char  u8;
typedef unsigned short u16;
typedef unsigned int   u32;

void *pMapmemory(off_t phy_addr, size_t phy_lenght);

volatile u32 *pGPIO_BASE;

// look at PXA255 Developer's Manual page 4-44
#define GPLR2 (*(pGPIO_BASE + 0x08 / 4))	/* GPLR2 @ 0x40E00008 */
#define GPDR2 (*(pGPIO_BASE + 0x14 / 4))	/* GPDR2 @ 0x40E00014 */
#define GPSR2 (*(pGPIO_BASE + 0x20 / 4))	/* GPSR2 @ 0x40E00020 */
#define GPCR2 (*(pGPIO_BASE + 0x2C / 4))	/* GPCR2 @ 0x40E0002C */

void write_portc(u8 data)
{
u32 gpdata2;

    /* GPIO84 = port C[0] */
    /* GPIO83 = port C[1] */
    /* GPIO81 = port C[2] */
    /* GPIO82 = port C[3] */
    gpdata2 = (((data & 1) << (20-0)) | ((data & 2) << (19-1)) |
	       ((data & 4) << (17-2)) | ((data & 8) << (18-3)));

    // clear port C
    GPCR2 = gpdata2^(0xf << 17);
    // set port C
    GPSR2 = gpdata2;
}

int main (void)
{
    u8 cValue;

    // check user identity
    if (geteuid() != 0) {
	printf("No root access rights !\n");
	exit(1);
    }

    // map PXA255 GPIO register space at 0x40E00000 / size 4096 bytes
    pGPIO_BASE = pMapmemory(0x40E00000, 0x00001000);
    if (pGPIO_BASE == NULL) {
	printf("Can't map memory at 0x40E00000 !\n");
	exit(1);
    }

    // set DIL/NetPC PIO port C = output
    GPDR2 |= (0x0f << 17);

    // run until user break by CTRL-C...
    cValue = 0;
    for(;;) {
	    // write value to port C
	    write_portc(cValue);
	    // print value port C
	    printf ("\r  Current value on port C = 0x%1X", cValue & 0xf);
	    fflush(stdout);

	    cValue++;
	    usleep (250000);
    }
    return(0);
}

void *pMapmemory(off_t phy_addr, size_t phy_lenght)
{
#define MAP_PAGESIZE 4096UL

int iFd;
void *pMem;

    if ((phy_addr % MAP_PAGESIZE) != 0) {
	fprintf(stderr,"physical address not aligned on PAGESIZE boundary!\n");
	return(NULL);
    }

    if ((phy_lenght % MAP_PAGESIZE) != 0) {
	fprintf(stderr,"physical lenght not aligned on PAGESIZE boundary!\n");
	return(NULL);
    }

    // open mem device for read/write
    iFd = open("/dev/mem", O_RDWR | O_SYNC);
    if (iFd < 0) {
	fprintf(stderr,"open of /dev/mem fail !\n");
	return(NULL);
    }

    // get pointer to DNP1110 memory
    pMem = mmap(NULL,
		phy_lenght,
		(PROT_READ | PROT_WRITE),
		MAP_SHARED,
		iFd,
		phy_addr);

    if ((pMem == MAP_FAILED) || (pMem == NULL)) {
	fprintf(stderr,"mmap of /dev/mem fail !\n");
	return(NULL);
    }

    // close mem device
    if (close(iFd) != 0)
	fprintf(stderr,"close of /dev/mem fail !\n");

    return(pMem);
}
